import atexit
import collections
import threading
import traceback
from decimal import Decimal

import requests

from werkzeug.exceptions import BadRequest

from zish import loads

from chellow.models import Contract, RateScript, Session, get_non_core_contract_id
from chellow.utils import (
    HH,
    c_months_u,
    ct_datetime,
    hh_format,
    to_ct,
    to_utc,
    utc_datetime_now,
)


def hh(data_source, provider="APXMIDP"):
    try:
        cache = data_source.caches["bmarketidx"]
    except KeyError:
        cache = data_source.caches["bmarketidx"] = {}

    for hh in data_source.hh_data:
        try:
            hh["bmarketidx-rate"] = cache[hh["start-date"]][provider]
        except KeyError:
            h_start = hh["start-date"]
            db_id = get_non_core_contract_id("bmarketidx")
            rates = data_source.hh_rate(db_id, h_start)["rates"]

            try:
                idxs = cache[h_start]
            except KeyError:
                idxs = cache[h_start] = {}

            try:
                rate = rates[h_start]
            except KeyError:
                try:
                    rate = sorted(rates.items())[-1][1]
                except KeyError:
                    raise BadRequest(
                        f"For the bmarketidx rate script at "
                        f"{hh_format(h_start)} the rate cannot be found."
                    )

            try:
                idx = rate[provider]
            except KeyError:
                try:
                    idx = sorted(rate.items())[0][1]
                except KeyError:
                    raise BadRequest(
                        f"For the bmarketidx rate script at "
                        f"{hh_format(h_start)} a rate cannot be found for the "
                        f"provider {provider}."
                    )

            hh["bmarketidx-rate"] = idxs[provider] = float(idx)


bmarketidx_importer = None


class BmarketidxImporter(threading.Thread):
    def __init__(self):
        super().__init__(name="Bmarketidx Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = self.global_alert = None
                try:
                    sess = Session()
                    self.log("Starting to check bmarketidx.")
                    contract = Contract.get_non_core_by_name(sess, "bmarketidx")
                    latest_rs = (
                        sess.query(RateScript)
                        .filter(RateScript.contract_id == contract.id)
                        .order_by(RateScript.start_date.desc())
                        .first()
                    )
                    start_ct = to_ct(latest_rs.start_date)

                    months = list(
                        c_months_u(
                            start_year=start_ct.year,
                            start_month=start_ct.month,
                            months=2,
                        )
                    )
                    month_start, month_finish = months[1]

                    now = utc_datetime_now()
                    if now > month_finish:
                        _process_month(
                            self.log,
                            sess,
                            contract,
                            latest_rs,
                            month_start,
                            month_finish,
                        )

                except BaseException:
                    self.log(f"Outer problem {traceback.format_exc()}")
                    sess.rollback()
                    self.global_alert = (
                        "There's a problem with the " "bmarketidx automatic importer."
                    )
                finally:
                    self.lock.release()
                    self.log("Finished checking bmarketidx rates.")
                    if sess is not None:
                        sess.close()

            self.going.wait(2 * 60 * 60)
            self.going.clear()


def _process_month(log_f, sess, contract, latest_rs, month_start, month_finish):
    latest_rs_id = latest_rs.id
    log_f(
        f"Checking to see if data is available from {hh_format(month_start)} "
        f"to {hh_format(month_finish)} on BMRS."
    )
    rates = {}
    month_finish_ct = to_ct(month_finish)
    for d in range(month_finish_ct.day):
        day_ct = ct_datetime(month_finish_ct.year, month_finish_ct.month, d + 1)
        params = {
            "q": f"ajax/alldata/MID/Date,SP,Provider,Price,Volume/NULL/"
            f'{day_ct.strftime("%Y-%m-%d")}/ALL'
        }
        r = requests.get(
            "https://www.bmreports.com/bmrs/", params=params, timeout=60, verify=False
        )
        res = r.json()
        for h in res["arr"]:
            dt = to_utc(day_ct + (int(h["settlementPeriod"]) - 1) * HH)
            try:
                rate = rates[dt]
            except KeyError:
                rate = rates[dt] = {}
            rate[h["DataProviderId"]] = Decimal(h["MarketIndexPrice"]) / Decimal(1000)

    if month_finish in rates:
        log_f("The whole month's data is there.")
        script = {"rates": rates}
        rs = RateScript.get_by_id(sess, latest_rs_id)
        contract.update_rate_script(
            sess, rs, rs.start_date, month_finish, loads(rs.script)
        )
        contract.insert_rate_script(sess, month_start, script)
        sess.commit()
        log_f(f"Added a new rate script starting at {hh_format(month_start)}.")
    else:
        msg = "There isn't a whole month there yet."
        if len(rates) > 0:
            msg += " The last date is {sorted(rates.keys())[-1]}"
        log_f(msg)


def get_importer():
    return bmarketidx_importer


def startup():
    global bmarketidx_importer
    bmarketidx_importer = BmarketidxImporter()
    bmarketidx_importer.start()


@atexit.register
def shutdown():
    if bmarketidx_importer is not None:
        bmarketidx_importer.stop()



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.edi_lib import EdiParser, SEGMENTS
from chellow.utils import HH, to_ct, to_utc


read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}


TCOD_MAP = {
    "030025": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "345065": ("summer-weekend-gbp", "summer-weekend-rate", "summer-weekend-kwh"),
    "350293": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "425779": ("ro-gbp", "ro-rate", "ro-kwh"),
    "534342": ("reconciliation-gbp", None, None),
    "584867": ("aahedc-gbp", "aahedc-rate", "aahedc-kwh"),
    "989534": ("bsuos-gbp", "bsuos-rate", "bsuos-kwh"),
    "117220": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "579387": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "558147": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "066540": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "281170": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "342094": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "378809": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "574015": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "810016": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "839829": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "068476": ("day-gbp", "day-rate", "day-kwh"),
    "219182": (
        "duos-availability-gbp",
        "duos-availability-rate",
        "duos-availability-kva",
    ),
    "144424": (
        "duos-excess-availability-gbp",
        "duos-excess-availability-rate",
        "duos-excess-availability-kva",
    ),
    "986159": ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
    "838286": ("duos-reactive-gbp", "duos-reactive-rate", "duos-reactive-kvarh"),
    "242643": ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
    "257304": ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh"),
    "661440": ("duos-amber-gbp", "duos-amber-rate", "duos-red-kwh"),
    "257305": ("duos-green-gbp", "duos-green-rate", "duos-green-kwh"),
    "257303": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "661439": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "661441": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "309707": ("fit-gbp", "fit-rate", "fit-kwh"),
    "994483": ("reconciliation-gbp", None, None),
    "544936": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "265091": ("night-gbp", "night-rate", "night-kwh"),
    "483457": ("peak-gbp", "peak-rate", "peak-kwh"),
    "975901": ("peak-shoulder-gbp", "peak-shoulder-rate", "peak-shoulder-kwh"),
    "364252": ("ro-gbp", "ro-rate", "ro-kwh"),
    "378246": ("ro-gbp", "ro-rate", "ro-kwh"),
    "632209": ("summer-night-gbp", "summer-night-rate", "summer-night-kwh"),
    "663682": ("summer-weekday-gbp", "summer-weekday-rate", "summer-weekday-kwh"),
    "299992": ("summer-weekend-gbp", "summer-weekend-rate", "summer-weekend-kwh"),
    "447769": ("triad-gbp", "triad-rate", "triad-kw"),
    "647721": ("triad-gbp", "triad-rate", "triad-kw"),
    "276631": ("triad-gbp", "triad-rate", "triad-kw"),
    "220894": ("winter-night-gbp", "winter-night-rate", "winter-night-kwh"),
    "264929": ("winter-weekday-gbp", "winter-weekday-rate", "winter-weekday-kwh"),
    "638187": ("winter-weekend-gbp", "winter-weekend-rate", "winter-weekend-kwh"),
    "700285": ("duo-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
}


def _to_date(component):
    return to_utc(to_ct(Datetime.strptime(component, "%y%m%d")))


def _to_decimal(components, divisor=None):
    comp_0 = components[0]

    try:
        result = Decimal(comp_0)
    except InvalidOperation as e:
        raise BadRequest(
            "Can't parse '"
            + str(comp_0)
            + "' of "
            + str(components)
            + " as a decimal:"
            + str(e)
        )

    if len(components) > 1 and components[-1] == "R":
        result *= Decimal("-1")

    if divisor is not None:
        result /= Decimal(divisor)

    return result


def _find_elements(code, elements):
    segment_name = code + elements[1][0] if code == "CCD" else code
    elem_codes = [m["code"] for m in SEGMENTS[segment_name]["elements"]]
    return dict(zip(elem_codes, elements))


class Parser:
    def __init__(self, f):
        self.parser = EdiParser(StringIO(str(f.read(), "utf-8", errors="ignore")))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        headers = {}
        for self.line_number, code in enumerate(self.parser):
            elements = _find_elements(code, self.parser.elements)
            line = self.parser.line
            try:
                bill = _process_segment(code, elements, line, headers)
            except BadRequest as e:
                raise BadRequest("Can't parse the line: " + line + " :" + e.description)
            if bill is not None:
                raw_bills.append(bill)
        return raw_bills


def _process_segment(code, elements, line, headers):
    if code == "BCD":
        issue_date = _to_date(elements["IVDT"][0])
        reference = elements["INVN"][0]
        bill_type_code = elements["BTCD"][0]

        sumo = elements["SUMO"]
        start_date = _to_date(sumo[0])
        finish_date = _to_date(sumo[1]) + relativedelta(days=1) - HH

        headers["issue_date"] = issue_date
        headers["bill_type_code"] = bill_type_code
        headers["reference"] = reference

    elif code == "MHD":
        message_type = elements["TYPE"][0]
        if message_type == "UTLBIL":
            pass

    elif code == "CCD":
        consumption_charge_indicator = elements["CCDE"][0]

        if consumption_charge_indicator == "1":
            pres_read_date = _to_date(elements["PRDT"][0]) + relativedelta(days=1) - HH

            prev_read_date = _to_date(elements["PVDT"][0]) + relativedelta(days=1) - HH

            m = elements["MLOC"][0]
            mpan = " ".join(
                (m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13])
            )

            prrd = elements["PRRD"]
            pres_read_type = read_type_map[prrd[1]]
            prev_read_type = read_type_map[prrd[3]]

            coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
            pres_reading_value = Decimal(prrd[0])
            prev_reading_value = Decimal(prrd[2])
            msn = elements["MTNR"][0]
            tpr_code = elements["TMOD"][0]
            if tpr_code == "kW":
                units = "kW"
                tpr_code = None
            elif tpr_code == "kVA":
                units = "kVA"
                tpr_code = None
            else:
                units = "kWh"
                kwh = _to_decimal(elements["CONS"], "1000")

            try:
                reads = headers["reads"]
            except KeyError:
                reads = headers["reads"] = []

            reads.append(
                {
                    "msn": msn,
                    "mpan": mpan,
                    "coefficient": coefficient,
                    "units": units,
                    "tpr_code": tpr_code,
                    "prev_date": prev_read_date,
                    "prev_value": prev_reading_value,
                    "prev_type_code": prev_read_type,
                    "pres_date": pres_read_date,
                    "pres_value": pres_reading_value,
                    "pres_type_code": pres_read_type,
                }
            )

        elif consumption_charge_indicator == "2":
            breakdown = defaultdict(int, {"raw-lines": line})

            element_code = elements["TCOD"][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code + " in the TCOD_MAP."
                )

            m = elements["MLOC"][0]
            mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements["CONS"]
            kwh = Decimal("0")
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, "1000")
                breakdown[eln_cons] = kwh = el_cons

            if eln_rate is not None:
                rate = _to_decimal(elements["BPRI"], "100000")
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements["CSDT"][0])
            headers["bill_start_date"] = start_date

            finish_date = _to_date(elements["CEDT"][0]) - HH
            headers["bill_finish_date"] = finish_date

            if "CTOT" in elements:
                net = Decimal("0.00") + _to_decimal(elements["CTOT"], "100")
            else:
                net = Decimal("0.00")

            breakdown[eln_gbp] = net

            headers["mpan_core"] = mpan_core

            try:
                reads = headers["reads"]
                del headers["reads"][:]
            except KeyError:
                reads = []

            return {
                "bill_type_code": headers["bill_type_code"],
                "reference": headers["reference"] + "_" + eln_gbp[:-4],
                "issue_date": headers["issue_date"],
                "mpan_core": mpan_core,
                "account": mpan_core,
                "start_date": start_date,
                "finish_date": finish_date,
                "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
                "net": net,
                "vat": Decimal("0.00"),
                "gross": net,
                "breakdown": breakdown,
                "reads": reads,
            }

        elif consumption_charge_indicator == "3":
            breakdown = defaultdict(int, {"raw-lines": line})

            element_code = elements["TCOD"][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code + " in the TCOD_MAP."
                )

            m = elements["MLOC"][0]
            mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements["CONS"]
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, "1000")
                breakdown[eln_cons] = kwh = el_cons
            else:
                kwh = Decimal("0")

            if eln_rate is not None:
                rate = _to_decimal(elements["BPRI"], "100000")
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements["CSDT"][0])
            headers["bill_start_date"] = start_date

            finish_date = _to_date(elements["CEDT"][0]) - HH
            headers["bill_finish_date"] = finish_date

            if "CTOT" in elements:
                net = Decimal("0.00") + _to_decimal(elements["CTOT"], "100")
            else:
                net = Decimal("0.00")

            breakdown[eln_gbp] = net

            headers["mpan_core"] = mpan_core

            try:
                reads = headers["reads"]
                del headers["reads"][:]
            except KeyError:
                reads = []

            return {
                "bill_type_code": headers["bill_type_code"],
                "issue_date": headers["issue_date"],
                "reference": headers["reference"],
                "mpan_core": mpan_core,
                "account": mpan_core,
                "start_date": start_date,
                "finish_date": finish_date,
                "net": net,
                "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
                "vat": Decimal("0.00"),
                "gross": net,
                "breakdown": breakdown,
                "reads": reads,
            }

        elif consumption_charge_indicator == "4":
            breakdown = defaultdict(int, {"raw-lines": line})

            element_code = elements["TCOD"][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code + " in the TCOD_MAP."
                )

            m = elements["MLOC"][0]
            mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements["CONS"]
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, "1000")
                breakdown[eln_cons] = kwh = el_cons

            if eln_rate is not None:
                rate = _to_decimal(elements["BPRI"], "100000")
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements["CSDT"][0])
            headers["bill_start_date"] = start_date

            finish_date = _to_date(elements["CEDT"][0]) - HH
            headers["bill_finish_date"] = finish_date

            if "CTOT" in elements:
                net = Decimal("0.00") + _to_decimal(elements["CTOT"], "100")
            else:
                net = Decimal("0.00")

            breakdown[eln_gbp] = net

            headers["mpan_core"] = mpan_core

            try:
                reads = headers["reads"]
                del headers["reads"][:]
            except KeyError:
                reads = []

            return {
                "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0.00"),
                "reference": headers["reference"] + "_" + eln_gbp[:-4],
                "issue_date": headers["issue_date"],
                "mpan_core": mpan_core,
                "account": mpan_core,
                "start_date": start_date,
                "finish_date": finish_date,
                "net": net,
                "vat": Decimal("0.00"),
                "gross": net,
                "breakdown": breakdown,
                "reads": reads,
                "bill_type_code": headers["bill_type_code"],
            }

    elif code == "MTR":
        headers.clear()

    elif code == "MAN":
        pass

    elif code == "VAT":
        vat = Decimal("0.00") + _to_decimal(elements["UVTT"], "100")

        return {
            "bill_type_code": headers["bill_type_code"],
            "account": headers["mpan_core"],
            "mpan_core": headers["mpan_core"],
            "reference": headers["reference"] + "_vat",
            "issue_date": headers["issue_date"],
            "start_date": headers["bill_start_date"],
            "finish_date": headers["bill_finish_date"],
            "kwh": Decimal("0.00"),
            "net": Decimal("0.00"),
            "vat": vat,
            "gross": vat,
            "breakdown": {},
            "reads": [],
        }


def _get_element(elements, index, line):
    try:
        return elements[index]
    except IndexError:
        raise BadRequest(
            "The index "
            + str(index)
            + " is past the end of the elements "
            + str(elements)
            + " and line "
            + line
            + "."
        )



import csv
import os
import threading
import traceback
from datetime import datetime as Datetime
from itertools import chain

from flask import g, request

from sqlalchemy import Float, cast, func, or_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.computer
import chellow.dloads
import chellow.duos
from chellow.models import (
    Bill,
    BillType,
    Channel,
    Era,
    HhDatum,
    Llfc,
    Mtc,
    ReadType,
    RegisterRead,
    Session,
    SiteEra,
    Supply,
)
from chellow.utils import HH, hh_format, hh_max, hh_min, req_hh_date, req_int, to_utc
from chellow.views import chellow_redirect


NORMAL_READ_TYPES = "C", "N", "N3"


def mpan_bit(
    sess,
    supply,
    is_import,
    num_hh,
    era,
    chunk_start,
    chunk_finish,
    forecast_date,
    caches,
):
    mpan_core_str = llfc_code = sc_str = supplier_contract_name = num_bad = ""
    gsp_kwh = msp_kwh = md = non_actual = 0
    date_at_md = kvarh_at_md = None

    mpan_core = era.imp_mpan_core if is_import else era.exp_mpan_core
    if mpan_core is not None:
        if num_bad == "":
            num_bad = 0

        mpan_core_str = mpan_core
        if is_import:
            supplier_contract_name = era.imp_supplier_contract.name
            llfc = era.imp_llfc
            sc = era.imp_sc
        else:
            supplier_contract_name = era.exp_supplier_contract.name
            llfc = era.exp_llfc
            sc = era.exp_sc
        llfc_code = llfc.code
        sc_str = str(sc)

        supply_source = chellow.computer.SupplySource(
            sess, chunk_start, chunk_finish, forecast_date, era, is_import, caches
        )

        if era.energisation_status.code == "E":
            chellow.duos.duos_vb(supply_source)
            for hh in supply_source.hh_data:
                gsp_kwh += hh["gsp-kwh"]
                hh_msp_kwh = hh["msp-kwh"]
                msp_kwh += hh_msp_kwh
                if hh["status"] != "A":
                    num_bad += 1
                    non_actual += hh_msp_kwh
                if hh_msp_kwh > md:
                    md = hh_msp_kwh
                    date_at_md = hh["start-date"]

    if date_at_md is not None:
        kvarh_at_md = (
            sess.query(cast(func.max(HhDatum.value), Float))
            .join(Channel)
            .filter(
                Channel.era == era,
                Channel.imp_related == is_import,
                Channel.channel_type != "ACTIVE",
                HhDatum.start_date == date_at_md,
            )
            .scalar()
        )

    kw_at_md = md * 2

    if kvarh_at_md is None:
        kva_at_md = "None"
    else:
        kva_at_md = (kw_at_md ** 2 + (kvarh_at_md * 2) ** 2) ** 0.5

    date_at_md_str = "" if date_at_md is None else hh_format(date_at_md)

    return [
        llfc_code,
        mpan_core_str,
        sc_str,
        supplier_contract_name,
        msp_kwh,
        non_actual,
        gsp_kwh,
        kw_at_md,
        date_at_md_str,
        kva_at_md,
        num_bad,
    ]


def content(supply_id, start_date, finish_date, user):
    forecast_date = to_utc(Datetime.max)
    caches = {}
    f = sess = era = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "supplies_duration.csv", user
        )
        f = open(running_name, mode="w", newline="")
        w = csv.writer(f, lineterminator="\n")
        w.writerow(
            (
                "Era Start",
                "Era Finish",
                "Supply Id",
                "Supply Name",
                "Source",
                "Generator Type",
                "Site Code",
                "Site Name",
                "Associated Site Codes",
                "From",
                "To",
                "PC",
                "MTC",
                "CoP",
                "SSC",
                "Energisation Status",
                "Properties",
                "MOP Contract",
                "MOP Account",
                "DC Contract",
                "DC Account",
                "Normal Reads",
                "Type",
                "Supply Start",
                "Supply Finish",
                "Import LLFC",
                "Import MPAN Core",
                "Import Supply Capacity",
                "Import Supplier",
                "Import Total MSP kWh",
                "Import Non-actual MSP kWh",
                "Import Total GSP kWh",
                "Import MD / kW",
                "Import MD Date",
                "Import MD / kVA",
                "Import Bad HHs",
                "Export LLFC",
                "Export MPAN Core",
                "Export Supply Capacity",
                "Export Supplier",
                "Export Total MSP kWh",
                "Export Non-actual MSP kWh",
                "Export GSP kWh",
                "Export MD / kW",
                "Export MD Date",
                "Export MD / kVA",
                "Export Bad HHs",
            )
        )

        eras = (
            sess.query(Era)
            .filter(
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
                Era.start_date <= finish_date,
            )
            .order_by(Era.supply_id, Era.start_date)
            .options(
                joinedload(Era.supply),
                joinedload(Era.supply).joinedload(Supply.source),
                joinedload(Era.supply).joinedload(Supply.generator_type),
                joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.imp_llfc),
                joinedload(Era.exp_llfc),
                joinedload(Era.mop_contract),
                joinedload(Era.dc_contract),
                joinedload(Era.channels),
                joinedload(Era.site_eras).joinedload(SiteEra.site),
                joinedload(Era.pc),
                joinedload(Era.cop),
                joinedload(Era.mtc).joinedload(Mtc.meter_type),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.ssc),
                joinedload(Era.energisation_status),
                joinedload(Era.site_eras),
            )
        )

        if supply_id is not None:
            eras = eras.filter(Era.supply == Supply.get_by_id(sess, supply_id))

        for era in eras:
            supply = era.supply
            site_codes = set()
            site = None
            for site_era in era.site_eras:
                if site_era.is_physical:
                    site = site_era.site
                else:
                    site_codes.add(site_era.site.code)

            sup_eras = (
                sess.query(Era)
                .filter(Era.supply == supply)
                .order_by(Era.start_date)
                .all()
            )
            supply_start = sup_eras[0].start_date
            supply_finish = sup_eras[-1].finish_date

            if supply.generator_type is None:
                generator_type = ""
            else:
                generator_type = supply.generator_type.code

            ssc = era.ssc
            ssc_code = "" if ssc is None else ssc.code

            prime_reads = set()
            for read, rdate in chain(
                sess.query(RegisterRead, RegisterRead.previous_date)
                .join(RegisterRead.previous_type)
                .join(Bill)
                .join(BillType)
                .filter(
                    Bill.supply == supply,
                    BillType.code != "W",
                    RegisterRead.previous_date >= start_date,
                    RegisterRead.previous_date <= finish_date,
                    ReadType.code.in_(NORMAL_READ_TYPES),
                )
                .options(joinedload(RegisterRead.bill)),
                sess.query(RegisterRead, RegisterRead.present_date)
                .join(RegisterRead.present_type)
                .join(Bill)
                .join(BillType)
                .filter(
                    Bill.supply == supply,
                    BillType.code != "W",
                    RegisterRead.present_date >= start_date,
                    RegisterRead.present_date <= finish_date,
                    ReadType.code.in_(NORMAL_READ_TYPES),
                )
                .options(joinedload(RegisterRead.bill)),
            ):
                prime_bill = (
                    sess.query(Bill)
                    .join(BillType)
                    .filter(
                        Bill.supply == supply,
                        Bill.start_date <= read.bill.finish_date,
                        Bill.finish_date >= read.bill.start_date,
                        Bill.reads.any(),
                    )
                    .order_by(Bill.issue_date.desc(), BillType.code)
                    .first()
                )
                if prime_bill.id == read.bill.id:
                    prime_reads.add(f"{rdate}_{read.msn}")

            supply_type = era.meter_category

            chunk_start = hh_max(era.start_date, start_date)
            chunk_finish = hh_min(era.finish_date, finish_date)
            num_hh = int((chunk_finish + HH - chunk_start).total_seconds() / (30 * 60))

            w.writerow(
                [
                    hh_format(era.start_date),
                    hh_format(era.finish_date, ongoing_str=""),
                    supply.id,
                    supply.name,
                    supply.source.code,
                    generator_type,
                    site.code,
                    site.name,
                    "| ".join(sorted(site_codes)),
                    hh_format(start_date),
                    hh_format(finish_date),
                    era.pc.code,
                    era.mtc.code,
                    era.cop.code,
                    ssc_code,
                    era.energisation_status.code,
                    era.properties,
                    era.mop_contract.name,
                    era.mop_account,
                    era.dc_contract.name,
                    era.dc_account,
                    len(prime_reads),
                    supply_type,
                    hh_format(supply_start),
                    hh_format(supply_finish, ongoing_str=""),
                ]
                + mpan_bit(
                    sess,
                    supply,
                    True,
                    num_hh,
                    era,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    caches,
                )
                + mpan_bit(
                    sess,
                    supply,
                    False,
                    num_hh,
                    era,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    caches,
                )
            )

            # Avoid a long-running transaction
            sess.rollback()
    except BadRequest as e:
        if era is None:
            pref = "Problem: "
        else:
            pref = (
                "Problem with era "
                + chellow.utils.url_root
                + "eras/"
                + str(era.id)
                + "/edit : "
            )
        f.write(pref + e.description)
    except BaseException as e:
        if era is None:
            pref = "Problem: "
        else:
            pref = "Problem with era " + str(era.id) + ": "
        f.write(pref + str(e))
        f.write(traceback.format_exc())
    finally:
        sess.close()
        f.close()
        os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_hh_date("start")
    finish_date = req_hh_date("finish")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    thread = threading.Thread(
        target=content, args=(supply_id, start_date, finish_date, g.user)
    )
    thread.start()
    return chellow_redirect("/downloads", 303)
